"""
Some common distributions on the sphere.
"""
import functools

import numpy as np
from scipy.special import i0, ive, logsumexp

from geosss import sphere
from geosss.utils import counter


class Distribution:
    def log_prob(self, x):
        pass


class Uniform(Distribution):
    def log_prob(self, x):
        return 0.0

    def gradient(self, x):
        return np.zeros_like(x)


@counter(["log_prob", "gradient"])
class Bingham(Distribution):
    """Bingham distribution

    p(x) \propto \exp(x^T A x)

    where A is a symmetric matrix without loss of generality.

    We can use a spectral decomposition to write
    \[
    A = U\Lambda U^T
    \]
    where $\Lambda$ is diagonal matrix of eigenvalues, and the columns of $U$
    are the eigenvectors of $A$. Moreover, $U$ is a column-orthogonal matrix:
    $U^T U = I$.

    Variable transformation:

    x -> y = U^Tx, x = Uy

    The transformed distribution is

    y \sim \exp(\sum_i \lambda_i y_i^2) = \prod_i \exp(\lambda_i y_i^2)

    subject to $\|y\| = 1$

    \exp(x^T A x) : Bingham distribution
    \exp(x^T b) : von Mises - Fisher distribution
    \exp(x^T A x + x^T b): Fisher-Bingham distribution
    """

    def __init__(self, A):
        assert np.allclose(A, A.T)

        self.A = A

        # eigh returns eigenvalues/-vectors in ascending order
        v, U = np.linalg.eigh(A)

        # descending order
        self.v, self.U = v[::-1], U[:, ::-1]

    def log_prob(self, x):
        """
        Evaluate the log pdf of the Bingham distribution for a single point on
        the sphere (if np.ndim(x) == 1) or multiple points (rows of a rank-2
        array).
        """
        assert x.ndim in (1, 2) and x.shape[-1] == self.d

        return np.sum((x @ self.A) * x, axis=-1)

    def gradient(self, x):
        return 2 * self.A @ x

    @property
    def d(self):
        """Dimension of ambient space."""
        return len(self.A)

    @property
    def mode(self):
        """Point with maximum probability."""
        return self.U[:, 0]

    @property
    def max_log_prob(self):
        return self.v[0]


@counter(["log_prob", "gradient"])
class BinghamFisher(Bingham):
    def __init__(self, A, b):
        super().__init__(A)
        assert len(A) == len(b)
        self.b = b

    def log_prob(self, x):
        return super().log_prob(x) + x @ self.b


@counter(["log_prob", "gradient"])
class VonMisesFisher(Distribution):
    """Von Mises-Fisher distribution

    vMF(x) \propto etr(mu^T x)

    where x is restricted to the unit sphere
    """

    def __init__(self, mu):
        self.mu = np.array(mu)

    @property
    def d(self):
        return self.mu.size

    @property
    def kappa(self):
        return np.linalg.norm(self.mu)

    @property
    def mode(self):
        return sphere.radial_projection(self.mu)

    @property
    def max_log_prob(self):
        return self.kappa

    @functools.cached_property
    def log_Z(self):
        d = self.d
        kappa = np.linalg.norm(self.mu)
        log_Z = (
            (d / 2) * np.log(2 * np.pi)
            + np.log(ive(d / 2 - 1, kappa))
            - (d / 2 - 2) * np.log(kappa)
        )
        return log_Z

    def log_prob(self, x):
        return x @ self.mu - np.log(2 * np.pi) - np.log(i0(np.linalg.norm(self.mu)))

    def gradient(self, x):
        return self.mu


@counter("log_prob")
class MultivariateNormal(Distribution):
    def __init__(self, mu, C):
        self.mu = mu
        self.C = C
        self.invC = np.linalg.inv(C)

    def log_prob(self, x):
        return -0.5 * np.sum((x - self.mu) * ((x - self.mu) @ self.invC), axis=-1)


@counter("log_prob")
class ACG(MultivariateNormal):
    def __init__(self, C):
        super().__init__(np.zeros(len(C)), C)

    def log_prob(self, x):
        y = -2 * super().log_prob(x)
        return -np.log(y) * len(self.C) / 2


@counter(["log_prob", "gradient"])
class MixtureModel(Distribution):
    def __init__(self, components, weights=None):
        self.pdfs = components
        if weights is None:
            weights = np.ones(len(components))
        self.weights = np.array(weights)
        self.weights /= self.weights.sum()

    def log_prob(self, x):
        p = np.transpose([pdf.log_prob(x) for pdf in self.pdfs])
        p += np.log(self.weights)
        return logsumexp(p, axis=-1)

    def gradient(self, x):
        p = np.transpose([pdf.log_prob(x) for pdf in self.pdfs])
        p += np.log(self.weights)
        g = np.transpose([pdf.gradient(x) for pdf in self.pdfs])
        return np.sum(np.exp(p) * g, axis=1) / np.exp(logsumexp(p, axis=-1))


def random_bingham(d=2, vmax=None, vmin=None, eigensystem=False, seed=None):
    """
    Create a random Bingham distribution where
    Parameters
    ----------
    d : integer >= 2
        Dimension of ambient space
    vmax : float or None
        Optional maximum eigenvalue of the precision matrix.
    vmin : float or None
        Optional minimum eigenvalue of the precision matrix.
    eigensystem : bool
        Flag specifying if Bingham distribution has diagonal precision matrix
        (i.e. we are working in the eigenvasis of A) or not. (Default value:
        False)
    """
    rng = np.random.default_rng(seed)

    A = rng.standard_normal((d, d))
    A = A.T @ A
    v, U = np.linalg.eigh(A)
    if vmin is not None:
        v += vmin - v.min()
    if vmax is not None:
        v *= vmax / v.max()
    if eigensystem:
        U = np.eye(d)
    A = (U * v) @ U.T
    return Bingham(A)



"""
Some utility functions for sampling from the sphere. 
"""
import numpy as np

# projections


def radial_projection(x):
    """
    Radial projection of (d+1)-dimensional point(s) to d-sphere.
    """
    norm = np.linalg.norm(x, axis=-1) + 1e-100
    if x.ndim == 1:
        return x / norm
    else:
        return x / norm[:, None]


def orthogonal_projection(x, y):
    """
    Map point(s) `x` into orthogonal complement of point `y`.
    """
    normal = radial_projection(y)
    return x - (x @ normal) * normal


def spherical_projection(x, v):
    """
    Map point(s) `x` into the great subsphere with pole `v`.
    """
    return radial_projection(orthogonal_projection(x, v))


# sampling


def sample_sphere(d=2, size=None, seed=None):
    """
    Draw a random point from d-sphere by drawing a (d+1)-dimensional point from the
    standard normal distribution and mapping it to d-sphere.
    """
    rng = np.random.default_rng(seed)
    x = (
        rng.standard_normal(d + 1)
        if size is None
        else rng.standard_normal((size, d + 1))
    )
    return radial_projection(x)


def sample_subsphere(v, seed=None):
    """
    Sample uniformly from the great subsphere with pole `v`.
    """
    rng = np.random.default_rng(seed)
    return spherical_projection(rng.standard_normal(len(v)), v)


# distances


def distance(x, y):
    """
    Great circle distance (assuming x, y are on the sphere).
    """
    return np.arccos(np.clip(np.sum(x * y, axis=-1), -1, 1))


# transformation between Cartesian and polar coordinates


def cartesian2polar(x):
    """Map points on the unit sphere to polar coordinates."""
    return np.mod(np.arctan2(x[:, 1], x[:, 0]), 2 * np.pi)


def polar2cartesian(theta):
    return np.transpose([np.cos(theta), np.sin(theta)])


def spherical2cartesian(phi, theta):
    return np.transpose(
        [np.cos(phi) * np.sin(theta), np.sin(phi) * np.sin(theta), np.cos(theta)]
    )


def cartesian2spherical(x):
    return cartesian2polar(x), np.mod(np.arccos(x[:, 2]), np.pi)


def sample_marginal(d, size=None, seed=None):
    rng = np.random.default_rng(seed)
    s = rng.beta(0.5, 0.5 * (d - 1), size=size)
    t = np.sqrt(s) * rng.choice([-1, 1], size=size)
    return t


def wrap(x, u, v):
    """See Mardia and Jupp for 'wrapping approach'."""
    theta = np.linalg.norm(x)
    return np.sin(theta) * u + np.cos(theta) * v


def slerp(u, v):
    theta = np.arccos(u @ v)

    def interpolation(phi, u=u, v=v, theta=theta):
        return (np.sin(phi) * v + np.sin(theta - phi) * u) / np.sin(theta)

    return interpolation


def givens(u, v, x):
    """
    Returns function that rotates `x` in the plane spanned by `u`, `v` by an
    angle that will be the argument of the function.
    """

    def rotate(theta, u=u, v=v, x=x):
        ux = u @ x
        vx = v @ x
        return (
            x
            + (np.cos(theta) - 1) * (ux * u + vx * v)
            + np.sin(theta) * (ux * v - vx * u)
        )

    return rotate



# preliminary implementation for variants of slice samplers

import numpy as np

from . import sphere

# some nice colors
colors = [
    (0.85, 0.3, 0.1),
    (0.15, 0.35, 0.6),
    (0.95, 0.7, 0.1),
    (0.0, 0.0, 0.0),
    (0.8, 0.8, 0.8),
]


def sample_slice(log_prob, current, bounds, return_n_reject=False):
    """
    Bracketing procedure as presented in the paper.
    """
    threshold = log_prob(current) + np.log(np.random.rand())
    L, R = bounds
    x = np.random.uniform(0, 2 * np.pi)
    L, R = x - 2 * np.pi, x
    n_reject = 0

    while True:
        x = np.random.uniform(L, R)
        if log_prob(x) > threshold:
            if return_n_reject:
                return x, n_reject
            else:
                return x
        elif x < current:
            L = x
        else:
            R = x
        n_reject += 1


def sample_slice2(log_prob, current, bounds):
    """
    Alternative version in which the brackets are first shifted.
    """
    threshold = log_prob(current) + np.log(np.random.rand())
    x = np.random.uniform(*bounds)
    L, R = x + bounds[0], x + bounds[1]

    while True:
        if log_prob(x) > threshold:
            return x
        elif x < current:
            L = x
        else:
            R = x
        x = np.random.uniform(L, R)


def metropolis_hastings(
    pdf, n_samples, stepsize=0.5, adapt_stepsize=False, reproject=True
):
    """
    Sample hyperspherical distribution with Metropolis Hastings.
    """
    samples = []
    x = sphere.sample_sphere(pdf.d - 1)
    n_acc = 0

    while len(samples) < n_samples:
        # propose
        r = np.sqrt(np.random.gamma(0.5 * len(x)) / 0.5) if reproject else 1.0
        y = sphere.radial_projection(r * x + stepsize * np.random.randn(pdf.d))

        # accept / reject
        accept = np.log(np.random.rand()) < pdf.log_prob(y) - pdf.log_prob(x)
        x = y if accept else x

        # adapt stepsize
        if adapt_stepsize:
            stepsize *= 1.02 if accept else 0.98

        samples.append(np.copy(x))
        n_acc += int(accept)

    print(f"stepsize: {stepsize}, acceptance rate: {(n_acc/n_samples):.1%}")

    return np.array(samples)


def slice_sampling(pdf, n_samples=1e3):
    """
    Slice sampling on the sphere using first bracketing procedure.
    """
    bounds = (-np.pi, np.pi)

    # initialization
    v = sphere.sample_sphere(pdf.d - 1)
    samples = [v]
    n_reject = []

    while len(samples) < n_samples:
        # draw uniformly from subsphere
        u = sphere.sample_subsphere(v)

        # define probability on slice
        def log_prob(theta, pdf=pdf, v=v, u=u):
            return pdf.log_prob(np.cos(theta) * v + np.sin(theta) * u)

        # slice sampling using first bracketing procedure starting at current
        # sample (corresponding to theta=0)
        theta, n = sample_slice(log_prob, 0.0, bounds, return_n_reject=True)

        v = np.cos(theta) * v + np.sin(theta) * u

        samples.append(v)
        n_reject.append(n)

    print(f"average number of rejections: {np.mean(n_reject)}")

    return np.array(samples)


def slice_sampling2(pdf, n_samples=1e3):
    """
    Slice sampling on the sphere using second bracketing procedure.
    """
    bounds = (-np.pi, np.pi)

    # initialization
    v = sphere.sample_sphere(pdf.d - 1)
    samples = [v]
    n_reject = []

    while len(samples) < n_samples:
        # draw uniformly from subsphere
        u = sphere.sample_subsphere(v)

        # define probability on slice
        def log_prob(theta, pdf=pdf, v=v, u=u):
            return pdf.log_prob(np.cos(theta) * v + np.sin(theta) * u)

        # slice sampling using second bracketing procedure starting at current
        # sample (corresponding to theta=0)
        theta, n = sample_slice2(log_prob, 0.0, bounds, return_n_reject=True)

        v = np.cos(theta) * v + np.sin(theta) * u

        samples.append(v)
        n_reject.append(n)

    print(f"average number of rejections: {np.mean(n_reject)}")

    return np.array(samples)


def slice_sampling3(pdf, n_samples=1e3):
    """
    Slice sampling on the sphere using slerp and bracketing.
    """
    # initialization
    v = sphere.sample_sphere(pdf.d - 1)

    samples = [v]
    n_reject = []

    while len(samples) < n_samples:
        # draw uniformly from sphere
        u = sphere.sample_sphere(pdf.d - 1)

        # slerp
        slerp = sphere.slerp(u, v)
        theta = np.arccos(u @ v)  # - 0.5 * np.pi

        bounds = (theta - np.pi, theta + np.pi)

        # define probability on slice
        def log_prob(theta, pdf=pdf, slerp=slerp):
            return pdf.log_prob(slerp(theta))

        # slice sampling using first bracketing procedure starting at current
        # sample (corresponding to theta=0)
        theta, n = sample_slice(log_prob, theta, bounds, return_n_reject=True)

        v = slerp(theta)

        samples.append(v)
        n_reject.append(n)

    print(f"average number of rejections: {np.mean(n_reject)}")

    return np.array(samples)


def slice_sampling4(pdf, n_samples=1e3):
    """
    Slice sampling on the sphere using random Givens rotations and bracketing.
    """
    bounds = (-np.pi, np.pi)

    # initialization
    x = sphere.sample_sphere(pdf.d - 1)

    samples = [x]
    n_reject = []

    while len(samples) < n_samples:
        # draw uniformly from sphere
        u = sphere.sample_sphere(pdf.d - 1)
        v = sphere.sample_sphere(pdf.d - 1)  # generates a small circle?

        # how to construct a great circle?
        v = x  # also generates a small circle
        # unless
        u = sphere.spherical_projection(u, x)

        # Givens rotation
        givens = sphere.givens(u, v, x)

        # define probability on slice
        def log_prob(theta, pdf=pdf, givens=givens):
            return pdf.log_prob(givens(theta))

        assert np.isclose(log_prob(0.0), pdf.log_prob(x))

        # slice sampling using first bracketing procedure starting at current
        # sample (corresponding to theta=0)
        theta, n = sample_slice(log_prob, 0.0, bounds, return_n_reject=True)

        x = givens(theta)

        samples.append(x)
        n_reject.append(n)

    print(f"average number of rejections: {np.mean(n_reject)}")

    return np.array(samples)


def spherical_rejection_sampling(pdf, n_samples=1e3):
    """
    Use rejection sampling to sample from slice.
    """
    bounds = (-np.pi, np.pi)
    bounds = (0, 2 * np.pi)

    # initialization
    v = sphere.sample_sphere(pdf.d - 1)
    samples = [v]

    n_reject = []
    while len(samples) < n_samples:
        # draw uniformly from subsphere
        u = sphere.sample_subsphere(v)

        # define probability on slice
        def log_prob(theta, pdf=pdf, v=v, u=u):
            return pdf.log_prob(np.cos(theta) * v + np.sin(theta) * u)

        # rejection sampling where current sample defines threshold
        threshold = log_prob(0.0) + np.log(np.random.rand())
        n = 0
        while True:
            theta = np.random.uniform(*bounds)
            if log_prob(theta) > threshold:
                n_reject.append(n)
                break
            n += 1

        v = np.cos(theta) * v + np.sin(theta) * u

        samples.append(v)

    print(f"average number of rejections: {np.mean(n_reject)}")

    return np.array(samples)



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Normalize

from geosss.distributions import Distribution

METHODS = ("sss-reject", "sss-shrink", "rwmh", "hmc")
ALGOS = {
    "sss-reject": "geoSSS (reject)",
    "sss-shrink": "geoSSS (shrink)",
    "rwmh": "RWMH",
    "hmc": "HMC",
}


def sphere_pdf(n_grid: int, pdf: Distribution):
    """spherical grid with pdf"""

    # spherical grid
    u = np.linspace(0, np.pi, n_grid)
    v = np.linspace(0, 2 * np.pi, n_grid)

    u_grid, v_grid = np.meshgrid(u, v)
    vertices = np.stack(
        [
            np.cos(v_grid) * np.sin(u_grid),
            np.sin(v_grid) * np.sin(u_grid),
            np.cos(u_grid),
        ],
        axis=2,
    )

    # spherical to cartesian
    sph2cart = (
        np.outer(np.cos(v), np.sin(u)),
        np.outer(np.sin(v), np.sin(u)),
        np.outer(np.ones_like(u), np.cos(u)),
    )

    # pdf values for the grid
    pdf_vals = np.array([np.exp(pdf.log_prob(val)) for val in vertices])

    return sph2cart, pdf_vals


def compare_samplers_3d(
    pdf: Distribution,
    samples: dict,
    n_grid: int = 100,
):
    for method in METHODS:
        msg = "visualization accepts only 3 dimension"
        assert samples[method].shape[1] == 3, msg

    sph2cart, pdf_vals = sphere_pdf(n_grid, pdf)
    pdfnorm = Normalize(vmin=pdf_vals.min(), vmax=pdf_vals.max())

    fig, axes = plt.subplots(
        2, 2, figsize=(8, 8), subplot_kw={"projection": "3d"}, sharex=True, sharey=True
    )

    for ax, method in zip(axes.flat, METHODS):
        ax.computed_zorder = False
        ax.plot_surface(
            *sph2cart, facecolors=plt.cm.terrain_r(pdfnorm(pdf_vals)), alpha=1, zorder=1
        )

        x = samples[method]
        ax.scatter(*x.T, c="tab:red", s=1, alpha=0.4, zorder=2)
        ax.set_title(ALGOS[method], pad=-30)
        ax.set_aspect("equal")
        ax.view_init(-140, 150)
        ax.axis("off")

    fig.tight_layout()
    plt.show()



from .demo_vis import compare_samplers_3d, sphere_pdf
from .distributions import (
    ACG,
    Bingham,
    BinghamFisher,
    Distribution,
    MixtureModel,
    MultivariateNormal,
    VonMisesFisher,
    random_bingham,
)
from .mcmc import (
    MetropolisHastings,
    RejectionSphericalSliceSampler,
    ShrinkageSphericalSliceSampler,
    SphericalHMC,
)
from .rand import sample_bingham, sample_bingham_2d, sample_bingham_3d, sample_vMF
from .sphere import (
    cartesian2polar,
    cartesian2spherical,
    distance,
    givens,
    orthogonal_projection,
    polar2cartesian,
    radial_projection,
    sample_sphere,
    sample_subsphere,
    spherical2cartesian,
    spherical_projection,
)
from .utils import IAT, SamplerLauncher, acf, acf_fft, colors, counter, n_eff, take_time



"""
Special-purpose random number generators for the von Mises-Fisher and the Bingham
distribution.
"""
import numpy as np
from scipy.optimize import brentq

from geosss import distributions as dist
from geosss import sphere


def rotate_north(u):
    """
    Find rotation of u onto north pole [0 ... 0 1].
    """
    v = np.zeros_like(u)
    v[-1] = 1

    U, _, V = np.linalg.svd(np.multiply.outer(v, u))

    R = U @ V

    if np.linalg.det(R) < 0.0:
        U[:, -1] *= -1
        R = U @ V

    return R


def sample_vMF(pdf, size=1):
    """
    Generates a random sample from the von Mises-Fisher distribution using the
    algorithm proposed by Wood (1994).
    """
    assert isinstance(pdf, dist.VonMisesFisher)

    if size > 1:
        return np.array([sample_vMF(pdf) for _ in range(int(size))])

    p = pdf.d - 1
    kappa = np.linalg.norm(pdf.mu)

    if np.isclose(kappa, 0.0):
        return sphere.sample_sphere(p)

    u = pdf.mu / kappa

    b0 = (-2 * kappa + (4 * kappa**2 + p**2) ** 0.5) / p
    x0 = (1 - b0) / (1 + b0)
    n = kappa * x0 + p * np.log(1 - x0**2)

    while True:
        Z = np.random.beta(0.5 * p, 0.5 * p)
        U = np.log(np.random.rand())
        W = (1 - (1 + b0) * Z) / (1 - (1 - b0) * Z)

        if (kappa * W + p * np.log(1 - x0 * W) - n) >= U:
            break

    # sample from d-2 sphere
    v = sphere.sample_sphere(p - 1)
    x = np.append((1 - W**2) ** 0.5 * v, W)
    R = rotate_north(u).T

    return R @ x


def sample_bingham_2d(pdf, n_samples=1):
    """Sample from 2D Bingham distribution.

    Sample from the 2D Bingham distribution by transforming it to a zero-
    centered von Mises distribution.
    """
    if not (isinstance(pdf, dist.Bingham) and pdf.d == 2):
        raise ValueError("expected 2D Bingham distribution")

    u = np.random.choice([0, 1], size=int(n_samples))
    kappa = 0.5 * (pdf.v[0] - pdf.v[1])
    phi = np.random.vonmises(0.0, kappa, int(n_samples))

    # von Mises is defined over [-pi, pi), we want phi to be defined over
    # [0, 2pi) therefore we need to shift it by pi
    # theta = 0.5 * np.squeeze(phi + (2 * u + 1) * np.pi)
    theta = 0.5 * np.squeeze(phi + 2 * u * np.pi)

    # Cartesian coordinates
    x = sphere.polar2cartesian(theta)

    return x @ pdf.U.T


def sample_bingham_3d(pdf, n_samples=1):
    """Sample from the 3D Bingham distribution by using a Gibbs sampler."""
    if not isinstance(pdf, dist.Bingham) or pdf.d != 3:
        raise ValueError("3D Bingham expected")

    v = pdf.v

    # initialization
    x = np.random.uniform(-1.0, 1)
    theta = np.arccos(x)

    samples = []
    while len(samples) < n_samples:
        # sample azimuthal angle phi
        kappa = np.sin(theta) ** 2 * 0.5 * (v[0] - v[1])
        u = np.random.choice([0, 1])
        # TODO: phi = 0.5 * (np.random.vonmises(0., kappa) + (2*u + 1) * np.pi)
        phi = 0.5 * (np.random.vonmises(0.0, kappa) + 2 * u * np.pi)

        # sample polar angle theta
        kappa = v[2] + (v[1] - v[0]) * np.cos(phi) ** 2 - v[1]
        C = 0.0 if kappa < 0 else kappa

        # rejection sampling of x=cos(theta)
        while True:
            x = np.random.uniform(-1.0, 1.0)
            u = np.random.rand()
            if np.log(u) < kappa * x**2 - C:
                theta = np.arccos(x)
                break

        samples.append((phi, theta))

    # map to cartesian coordinates
    x = sphere.spherical2cartesian(*np.transpose(samples))

    return x @ pdf.U.T


def bfind(vals):
    """
    `vals` are the eigenvalues of the precision matrix.
    """

    def fb(b):
        return 1 - np.sum(1 / (b + 2 * vals))

    if np.allclose(vals, 0.0):
        return len(vals)
    return brentq(fb, 1, len(vals))


def sample_bingham(A, n_samples, n_iter=1000, return_efficiency=False):
    """
    Implementation of rejection sampling algorithm by Kent, Ganeiber and
    Mardia.
    """
    if A.ndim == 1:
        v = -A
    elif A.ndim == 2:
        v, U = np.linalg.eigh(-A)

    v -= np.min(v)
    b = bfind(v)
    log_M = -(len(A) - b) / 2 + (len(A) / 2) * np.log(len(A) / b)

    n = n_samples
    efficiency = []
    samples = []
    for _ in range(n_iter):
        # sample from angular central Gaussian (ACG) proposal
        x = np.random.randn(n, len(A)) / np.sqrt(1 + 2 * v / b)
        x /= np.linalg.norm(x, axis=-1)[:, None]

        # evaluate unnormalized ACG, Bingham, and bound
        u = np.sum(np.square(x) * v, axis=-1)
        log_acg = -(len(A) / 2) * np.log(1 + 2 * u / b)
        log_bing = -u

        # acceptance probability
        log_prob = log_bing - log_acg - log_M

        # accept / reject
        mask = np.log(np.random.rand(n)) < log_prob
        n -= np.sum(mask)
        samples.append(x[mask])
        efficiency.append(np.mean(mask))
        if n <= 0:
            break

    samples = np.concatenate(samples, 0)[:n_samples]
    if A.ndim == 2:
        samples = samples @ U.T

    if False:
        print("mean efficiency:", np.mean(efficiency))

    if return_efficiency:
        return samples, np.mean(efficiency)

    return samples
